import datetime
import logging
import os
import time
import traceback
from typing import Any, Dict, List, Optional, Union, Tuple

import torch
import torch.distributed as dist
import torch.optim.lr_scheduler
import torch.utils.data as data
from allennlp import training
from allennlp.common import checks
from allennlp.common import util as common_util
from allennlp.common.file_utils import hardlink_or_copy
from allennlp.models import model
from allennlp.training import checkpointer
from allennlp.training import learning_rate_schedulers
from allennlp.training import momentum_schedulers
from allennlp.training import moving_average
from allennlp.training import util as training_util
from allennlp.nn.parallel import DdpWrappedModel
from overrides import overrides

logger = logging.getLogger(__name__)


@training.TrainerCallback.register("transfer_patience")
class TransferPatienceEpochCallback(training.TrainerCallback):

    def on_epoch(
        self,
        trainer: "GradientDescentTrainer",
        metrics: Dict[str, Any],
        epoch: int,
        is_primary: bool = True,
        **kwargs,
    ) -> None:
        if trainer._learning_rate_scheduler and trainer._learning_rate_scheduler.patience is not None:
            trainer._metric_tracker._patience = trainer._learning_rate_scheduler.patience
            trainer._metric_tracker._epochs_with_no_improvement = 0
        else:
            raise checks.ConfigurationError("Learning rate scheduler isn't properly setup!")


@training.Trainer.register("gradient_descent_validate_n", constructor="from_partial_objects")
class GradientDescentTrainer(training.GradientDescentTrainer):

    def __init__(self, model: model.Model, optimizer: torch.optim.Optimizer, data_loader: data.DataLoader,
                 patience: Optional[int] = None, validation_metric: Union[str, List[str]] = "-loss",
                 validation_data_loader: data.DataLoader = None, num_epochs: int = 20,
                 serialization_dir: Optional[str] = None, checkpointer: checkpointer.Checkpointer = None,
                 cuda_device: Optional[Union[int, torch.device]] = None, grad_norm: Optional[float] = None,
                 grad_clipping: Optional[float] = None,
                 learning_rate_scheduler: Optional[learning_rate_schedulers.LearningRateScheduler] = None,
                 momentum_scheduler: Optional[momentum_schedulers.MomentumScheduler] = None,
                 moving_average: Optional[moving_average.MovingAverage] = None,
                 callbacks: List[training.TrainerCallback] = None, distributed: bool = False, local_rank: int = 0,
                 world_size: int = 1, num_gradient_accumulation_steps: int = 1, use_amp: bool = False,
                 enable_default_callbacks: bool = True, run_confidence_checks: bool = True,
                 grad_scaling: bool = True,ddp_wrapped_model: Optional[DdpWrappedModel] = None) -> None:
        # TODO extract param to constructor (+ constructor method?)
        super().__init__(model, optimizer, data_loader, patience, validation_metric, validation_data_loader, num_epochs,
                         serialization_dir, checkpointer, cuda_device, grad_norm, grad_clipping,
                         learning_rate_scheduler, momentum_scheduler, moving_average, callbacks, distributed,
                         local_rank, world_size, num_gradient_accumulation_steps, use_amp,enable_default_callbacks,
                         run_confidence_checks,grad_scaling,ddp_wrapped_model)
        self.validate_every_n = 5




    @overrides
    def _try_train(self) -> Tuple[Dict[str, Any], int]:
        try:
            epoch_counter = self._maybe_restore_checkpoint()
        except RuntimeError:
            traceback.print_exc()
            raise checks.ConfigurationError(
                "Could not recover training from the checkpoint.  Did you mean to output to "
                "a different serialization directory or delete the existing serialization "
                "directory?"
            )

        training_util.enable_gradient_clipping(self.model, self._grad_clipping)

        logger.info("Beginning training.")

        val_metrics: Dict[str, float] = {}
        metrics: Dict[str, Any] = {}
        epochs_trained = 0
        training_start_time = time.time()

        metrics["best_epoch"] = self._metric_tracker.best_epoch
        for key, value in self._metric_tracker.best_epoch_metrics.items():
            metrics["best_validation_" + key] = value

        for epoch in range(self._num_epochs):
            epoch_start_time = time.time()
            train_metrics = self._train_epoch(epoch)

            if self._primary and self._checkpointer is not None:
                self._checkpointer.save_checkpoint(epoch, self)

            # Wait for the master to finish saving the model checkpoint
            if self._distributed:
                dist.barrier()

            # get peak of memory usage
            for key, value in train_metrics.items():
                if key.startswith("gpu_") and key.endswith("_memory_MB"):
                    metrics["peak_" + key] = max(metrics.get("peak_" + key, 0), value)
                elif key.startswith("worker_") and key.endswith("_memory_MB"):
                    metrics["peak_" + key] = max(metrics.get("peak_" + key, 0), value)

            this_epoch_val_metric: float = 0.0
            if self._validation_data_loader is not None:
                val_metrics = {}
                if epoch % self.validate_every_n == 0:
                    with torch.no_grad():
                        # We have a validation set, so compute all the metrics on it.
                        val_loss, val_reg_loss, num_batches = self._validation_loss(epoch)

                        # It is safe again to wait till the validation is done. This is
                        # important to get the metrics right.
                        if self._distributed:
                            dist.barrier()

                        val_metrics = training_util.get_metrics(
                            self.model,
                            val_loss,
                            val_reg_loss,
                            batch_loss=None,
                            batch_reg_loss=None,
                            num_batches=num_batches,
                            reset=True#,
                            #world_size=self._world_size,
                            #cuda_device=self.cuda_device,
                        )

                        # Check validation metric for early stopping
                        this_epoch_val_metric = self._metric_tracker.combined_score(val_metrics)
                        # self._metric_tracker.add_metric(this_epoch_val_metric)

                train_metrics["patience"] = self._metric_tracker._patience
                if self._metric_tracker.should_stop_early():
                    logger.info("Ran out of patience.  Stopping training.")
                    break

            # Create overall metrics dict
            training_elapsed_time = time.time() - training_start_time
            metrics["training_duration"] = str(datetime.timedelta(seconds=training_elapsed_time))
            metrics["training_start_epoch"] = epoch_counter
            metrics["training_epochs"] = epochs_trained
            metrics["epoch"] = epoch

            for key, value in train_metrics.items():
                metrics["training_" + key] = value
            for key, value in val_metrics.items():
                metrics["validation_" + key] = value

            if self._metric_tracker.is_best_so_far():
                # Update all the best_ metrics.
                # (Otherwise they just stay the same as they were.)
                metrics["best_epoch"] = epoch
                for key, value in val_metrics.items():
                    metrics["best_validation_" + key] = value

                self._metric_tracker.best_epoch_metrics = val_metrics

            if self._serialization_dir and self._primary:
                common_util.dump_metrics(
                    os.path.join(self._serialization_dir, f"metrics_epoch_{epoch}.json"),
                    metrics,
                )

            # The Scheduler API is agnostic to whether your schedule requires a validation metric -
            # if it doesn't, the validation metric passed here is ignored.
            if self._learning_rate_scheduler:
                self._learning_rate_scheduler.step(this_epoch_val_metric)
            if self._momentum_scheduler:
                self._momentum_scheduler.step(this_epoch_val_metric)

            if self._primary and self._checkpointer is not None:
                self._checkpointer.save_checkpoint(
                    epoch, self )

            # Wait for the master to finish saving the checkpoint
            if self._distributed:
                dist.barrier()

            if (
                self._should_validate_this_epoch
                and self._serialization_dir
                and self._metric_tracker.is_best_so_far()
            ):
                if self._ddp_wrapped_model is not None and self._ddp_wrapped_model.is_sharded:
                    # Each worker saves its own shard for now (we combine the shards later).
                    self._best_model_filename = os.path.join(
                        self._serialization_dir, f"best_w{self._rank}.th"
                    )
                else:
                    self._best_model_filename = os.path.join(self._serialization_dir, "best.th")

            # Wait for the primary process to finish saving the best
            if self._distributed:
                dist.barrier()

            for callback in self._callbacks:
                callback.on_epoch(self, metrics=metrics, epoch=epoch, is_primary=self._primary)

            epoch_elapsed_time = time.time() - epoch_start_time
            logger.info("Epoch duration: %s", datetime.timedelta(seconds=epoch_elapsed_time))

            if epoch < self._num_epochs - 1:
                time_per_epoch = training_elapsed_time / (
                        (epoch + 1) - self._start_after_epochs_completed
                )
                estimated_time_remaining = (
                    time_per_epoch * self._num_epochs
                ) - training_elapsed_time
                formatted_time = str(datetime.timedelta(seconds=int(estimated_time_remaining)))
                logger.info("Estimated training time remaining: %s", formatted_time)

            epochs_trained += 1
        else:
            epoch = self._num_epochs - 1

        if self._metric_tracker.is_best_so_far():
            logger.info(
                "Best validation performance so far. Copying weights to '%s/best.th'.",
                self._serialization_dir,
            )
            model_state, training_states = self.get_checkpoint_state()
            torch.save(model_state, os.path.join(self._serialization_dir, "best.th"))

        # Load the best model state before returning
        if self._best_model_filename is None or self._metric_tracker.is_best_so_far():
            self._finalize_model()
        else:
            # The model we're loading here has already been finalized.
            self._load_model_state(self._best_model_filename)

        return metrics, epoch
